/* eslint-env node */

'use strict'

// --------------------------------------------------------------------
// Imports
// --------------------------------------------------------------------

import { readFileSync } from 'node:fs'
import * as solace from './interpreter.mjs'

// --------------------------------------------------------------------
// Helpers
// --------------------------------------------------------------------

function removeShebang(source) {
  return source.slice(0, 2) === '#!' ? source.replace('#!', '//') : source
}

function parseArgs(args) {
  const filenames = []
  let benchmark = false
  let verbose = false

  for (let i = 0; i < args.length; ++i) {
    const arg = args[i]

    if (arg.startsWith('-')) { // flag, please?
      if (arg === '-b' || arg === '--benchmark') benchmark = true
      if (arg === '-v' || arg === '--verbose') verbose = true
      continue
    }

    filenames.push(arg)
  }

  return { filenames, benchmark, verbose }
}

// --------------------------------------------------------------------
// Main
// --------------------------------------------------------------------

/* eslint-disable no-console */

function compile(args) {
  const { filenames, benchmark, verbose } = parseArgs(args)

  const files = filenames.map(name => {
    // If benchmarking, read in all the files at once, so that we can just time the matching.
    return [name, benchmark ? removeShebang(readFileSync(name).toString()) : null]
  })

  const matchStartTime = Date.now()

  // Parsing -- bails out when the first error is encountered.
  const results = []
  const success = files.every(arr => {
    const source = arr[1] ?? removeShebang(readFileSync(arr[0]).toString())
    if (verbose) { console.error(arr[0]) }

    const result = solace.grammar.match(source, 'Program')
    if (!result.succeeded()) {
      console.error(arr[0] + ':\n' + result.message)
      return false
    }
    results.push(result)
    return true
  })

  if (!success) return null
  if (benchmark) console.error('Matching:', (Date.now() - matchStartTime) + 'ms')

  // Codegen
  const codegenStartTime = Date.now()
  const code = results.map(r => solace.semantics(r).toES6()).join('\n\n')
  if (benchmark) console.error('Codegen:', (Date.now() - codegenStartTime) + 'ms')

  return code
}

// Run as program
if (import.meta && process.argv[1]) {
  console.log(compile(process.argv.slice(2)))
}

// Imported as module
export default compile
