import fs from 'fs'
import { describe, it, expect, beforeAll } from 'vitest'
import { grammar as createGrammar } from 'ohm-js'

describe('Solace Grammar', () => {
  let grammar

  function check(input, shouldSucceed = true) {
    const match = grammar.match(input)
    const success = match.succeeded()

    if (shouldSucceed && !success) {
      console.log(`Failed to parse «${input}», error:`, match.message)
    } else if (!shouldSucceed && success) {
      console.log(`Invalid input successfully parsed: «${input}»`, match)
    }

    expect(success).toBe(shouldSucceed)
  }

  beforeAll(() => {
    const grammarText = fs.readFileSync('./src/grammar.ohm', 'utf-8')
    grammar = createGrammar(grammarText)
  })

  describe('Basic statements and expressions', () => {
    it('literals', () => {
      check('true')
      check('false')
      check('none')
      check('undefined')
    })
    it('integers', () => {
      check('42')
      check('0')
      check('1')
      check('023')
      check('9999')
    })
    it('integers with prefix', () => {
      check('-1')
      check('-23')
      check('-0')
      check('+0')
      check('+999')
    })
    it('exponents', () => {
      check('42e2') // 4200
      check('1e+3') // 1000
      check('2e-3') // 0.002
    })
    it('floats', () => {
      check('1.0')
      check('23.744')
      check('-0.1')
      check('+1.23')
      check('999.000001')
    })
    it('strings in doublequotes', () => {
      check('"Foo"')
      check('"23"')
      check('"Hello World"')
    })
    it('strings in singlequotes', () => {
      check("'Foo'")
      check("'23'")
      check("'Hello World'")
    })

    it('arithmetic expressions', () => {
      check("42 + 23")
      check("42 - 23")
      check("foo + 23 - 42e-2")
      check("foo + bar")
      check("42 - foo")
      check("bar - foo")
      check("bar-foo")
      check("bar-42")
      check("(42 - 23E2) + foo")
      check("42**2 * foo")
      check("(42 - 23) ** 2")
      check("(42 % 23) / bar*2")
    })
  })

  describe("Statements and Declarations", () => {
    it('if statements', () => {
      check(`if x > 5 { 10 }`)
      check(`if foo < bar { 1 } else { 2 }`)
      check(`
        if foo < 23 {
          "small"
        } else if foo > 42 {
          "large"
        } else {
          "medium"
        }
      `)
    })

    it('incomplete if statements', () => {
      check(`
        if > 23 {
          "small"
        }
      `, false)
      check(`
        if foo < 23 {
          "small"
        } else if {
          "wrong"
        }
      `, false)
      
    })

    it('basic functions and function calls', () => {
      check(`
        fn add(x,y) {
          return x + y
        }
        fn foo() {
          log(23)
        }

        add(2,3)
        foo()

        add(if answer is true { 19 } else { 0 }, 23)
      `)

      check(`fn () { return "wrong" }`) // anonymous function
    })

    it('variable declarations', () => {
      check(`
        var foo = 42
        const bar = 'Hello'
        live baz = (foo-1)*2
      `)
    })

    it('arrays', () => {
      check(`
        const arr = ["Pirates", "know", "how", "to", "arr!"]
        var ary = []
        ary.push(23)
        ary.push(5)
        ary[1] = 42
      `)
    })
  })

  describe("Expressions", () => {
    it('comparison', () => {
      check(`x == 5`)
      check(`x != 5`)
      check(`x is true`)
      check(`x <= 23`)
      check(`x >= 23`)
      check(`x > 23`)
      check(`x < 23`)
    })

    it('logic', () => {
      check(`x > 5 && !true && !(ary[0] > ary[1] || ary[0] == 5)`)
    })

    it('bitwise', () => {
      check(`x & y | (2 ^ 5)`)
    })

    it('if expressions', () => {
      check(`
        const y = if x > 5 {
          10
        } else {
          23
        }
      `)
      check(`const y = if true { 10 } else { 23 }`)

      // if expressions always need an else block
      check(`
        const y = if x > 5 {
          10
        }
      `, false)
    })

    it('nested if expressions', () => {
      check(`
        const y = if x > 5 {
          10
        } else if x < 2 {
          0
        } else {
          2
        }
      `)
      // if expressions always need an else block
      check(`
        const y = if x > 5 {
          10
        } else if x < 2 {
          0
        }
      `, false)
    })

    it('function expressions', () => {
      check(`const foo = fn () { return 42 }`) // anonymous function with fn
      check(`const bar = fn () => { return 3e2 }`) // anon func with fn and arrow
      check(`const baz = () => { return 3e2 }`) // anon arrow func
    })
  })

  describe("Example Program", () => {
    it('hello world', () => {
      check(`
        fn main() {
          const ary = ["foo", "bar"]
          ary[1] = 'World'
          console.log('Hello', ary[1])
        }
      `)
    })
  })
})

