import fs from 'fs'
import { grammar as createGrammar } from 'ohm-js'

function flattenIterNodes(nodes) {
  const result = []
  for (let i = 0; i < nodes.length; ++i) {
    if (nodes[i]._node.ctorName === '_iter') {
      result.push(...flattenIterNodes(nodes[i].children))
    } else {
      result.push(nodes[i])
    }
  }
  return result
}

// Comparison function for sorting nodes based on their interval's start index.
function compareByInterval(node, otherNode) {
  return node.source.startIdx - otherNode.source.startIdx
}

function nodeToES6(node, children) {
  const flatChildren = flattenIterNodes(children).sort(compareByInterval)

  // Keeps track of where the previous sibling ended, so that we can re-insert discarded
  // whitespace into the final output.
  let prevEndIdx = node.source.startIdx
  let code = ''

  for (let i = 0; i < flatChildren.length; ++i) {
    const child = flatChildren[i]

    // Restore any discarded whitespace between this node and the previous one.
    if (child.source.startIdx > prevEndIdx) {
      code += node.source.sourceString.slice(prevEndIdx, child.source.startIdx)
    }
    code += child.toES6()
    prevEndIdx = child.source.endIdx
  }
  return code
}

const grammarPath = `${import.meta.dirname}/grammar.ohm`
const grammarText = fs.readFileSync(grammarPath, 'utf-8')
const grammar = createGrammar(grammarText)
const semantics = grammar.createSemantics()

semantics.addOperation('toES6', {
  Program(sourceElements) {
    const { sourceString, startIdx, endIdx } = this.source
    return (
      sourceString.slice(0, startIdx) +
      nodeToES6(this, [sourceElements]) +
      sourceString.slice(endIdx)
    )
  },
  VariableDeclaration(varType, ident, _equals, expr) {
    const keyword = varType.sourceString
    const name = ident.toES6()
    const value = expr.toES6()

    switch (keyword) {
      case 'var': return `let ${name} = ${value}`
      case 'live': return `let ${name} = ${value} /* TODO: reactivity not yet implemented */`
      default: return `${keyword} ${name} = ${value}`
    }
  },
  IfStatement(_ifKw, expr, thenBlock, _elseKw, elseBlock) {
    const ifStmt = `if (${expr.toES6()}) ${thenBlock.toES6()}`
    const elseStmt = `${elseBlock.toES6()}`.trim()

    let output = ifStmt
    if (elseStmt.length) output += ` else ${elseStmt}`

    return output
  },
  EqualityExpression_eq(leftExpr, _eq, rightExpr) {
    return `${leftExpr.toES6()} === ${rightExpr.toES6()}`
  },
  IfExpression(_ifKw, expr, thenBlock, _elseKw, elseBlock) {
    const thenValue = thenBlock.blockAsExpression()
    const elseValue = elseBlock.blockAsExpression()
    return `(${expr.toES6()} ? ${thenValue} : ${elseValue})`
  },
  FunctionDeclaration(_fnKw, ident, _leftParen, params, _rightParen, block) {
    return `const ${ident.toES6()} = (${params.toES6()}) => ${block.toES6()}`
  },
  FunctionExpression_arrow(_maybeFnKw, _leftParen, params, _rightParen, _arrow, block) {
    return `(${params.toES6()}) => ${block.toES6()}`
  },
  FunctionExpression_anon(_fnKw, _leftParen, params, _rightParen, block) {
    return `(${params.toES6()}) => ${block.toES6()}`
  },
  _nonterminal(...children) {
    return nodeToES6(this, children)
  },
  _terminal() {
    return this.sourceString
  },
  _iter(...children) {
    return children.map(c => c.toES6()).join('')
  },
  _default(...children) {
    // Reconstruct source from children
    return children.map(child => {
      if (typeof child.toES6 === 'function') return child.toES6()
      return '' // Skip iteration nodes
    }).join(' ')
  },
})

// Blocks can be simple statements (if condition {/*block*/})
// or expressions (const foo = condition ? /*block*/ : /*block*/)
// which needs to return a value, so it is transpiled to an IIFE
// like (const foo = cond ? (() => { const x = 5; return x; })() : ... )
semantics.addOperation('blockAsExpression', {
  Block(_open, statementList, _close) {
    const statements = statementList.children
    if (statements.length === 0) return ''
    if (statements.length === 1) return statements[0].toES6()

    const transpiled = statements.map(stmt => stmt.toES6())
    const last = transpiled[transpiled.length - 1]
    const init = transpiled.slice(0, -1).join('; ')

    return `(() => { ${init}; return ${last}; })()`
  },
  IfExpression(_ifKw, expr, thenBlock, _elseKw, elseBlock) {
    const thenValue = thenBlock.blockAsExpression()
    const elseValue = elseBlock.blockAsExpression()
    return `(${expr.toES6()} ? ${thenValue} : ${elseValue})`
  },
})

semantics.addOperation('hoistDeclarations', {
  FunctionDeclaration(_fnKw, ident, _leftParen, _params, _rightParen, _block) {
    return new Map([[ident.sourceString, [ident.source]]])
  },
  FunctionExpression(_) {
    return new Map()
  },
  _iter: mergeBindings,
  _nonterminal: mergeBindings,
  _terminal() {
    return new Map()
  },
})


// Merge the bindings from the given `nodes` into a single map, where the value
// is an array of source locations that name is bound.
function mergeBindings(...nodes) {
  const bindings = new Map()
  for (const child of nodes.filter(c => !c.isLexical())) {
    child.hoistDeclarations().forEach((sources, ident) => {
      if (bindings.has(ident)) {
        bindings.get(ident).push(...sources) // Shadowed binding.
      } else {
        bindings.set(ident, sources) // Not shadowed at this level.
      }
    })
  }
  return bindings
}

export { grammar, semantics }
