import { describe, it, expect, beforeAll } from 'vitest'
import { grammar, semantics } from './interpreter.mjs'

function compile(source) {
  const match = grammar.match(source, 'Program')
  if (!match.succeeded()) {
    console.error(`Failed to parse «${source}», error:`, match.message)
    return null
  }
  return semantics(match).toES6()
}

describe('Solace Interpreter', () => {
  describe('Expressions', () => {
    it('addition/substraction', () => {
      expect(compile('a + b - c')).toBe('a + b - c')
    })
    it('multiplication/division', () => {
      expect(compile('a * b / c')).toBe('a * b / c')
    })
    it('mixed', () => {
      expect(compile('a + b * c')).toBe('a + b * c')
    })
    it('others', () => {
      expect(compile('a ** b % c')).toBe('a ** b % c')
    })
    it('negative', () => {
      expect(compile('-2 * c')).toBe('-2 * c')
    })
  })
  describe('Functions', () => {
    it('one-liner declaration', () => {
      const input = `fn foo() {}`
      const output = compile(input)
      expect(output).toBe('const foo = () => {}')
    })
    it('one-liner expression with fn', () => {
      const input = `const foo = fn () {}`
      const output = compile(input)
      expect(output).toBe('const foo = () => {}')
    })
    it('one-liner expression with arrow', () => {
      const input = `const foo = () => {}`
      const output = compile(input)
      expect(output).toBe('const foo = () => {}')
    })
    it('one-liner expression with fn and arrow', () => {
      const input = `const foo = fn () => {}`
      const output = compile(input)
      expect(output).toBe('const foo = () => {}')
    })
    it('multi-line declaration', () => {
      const input = `fn foo() {
        const x = 5
        return x
      }`
      const output = compile(input)
      expect(output).toBe(`const foo = () => {
        const x = 5
        return x
      }`)
    })
  })

  describe('Assignments', () => {
    it('var', () => {
      const input = 'var foo = 23'
      const output = compile(input)
      expect(output).toBe('let foo = 23')
    })
    it('let', () => {
      const input = 'let foo = 23'
      const output = compile(input)
      expect(output).toBe('let foo = 23')
    })
    it('const', () => {
      const input = 'const foo = 23'
      const output = compile(input)
      expect(output).toBe('const foo = 23')
    })
    it('array declaration', () => {
      const input = 'const foo = [1,2,3]'
      const output = compile(input)
      expect(output).toBe('const foo = [1,2,3]')
    })
    it('array usage', () => {
      const input = 'const foo = ary[2]'
      const output = compile(input)
      expect(output).toBe('const foo = ary[2]')
    })
  })

  describe('Conditions and Comparisons', () => {
    it('if', () => {
      const input = 'if x < 5 { console.log(x) }'
      const output = compile(input)
      expect(output).toBe('if (x < 5) { console.log(x) }')
    })
    it('if-else', () => {
      const input = 'if x < 5 { console.log(x) } else { console.log("no") }'
      const output = compile(input)
      expect(output).toBe('if (x < 5) { console.log(x) } else { console.log("no") }')
    })
    it('nested if', () => {
      const input = `if x < 5 {
        console.log('small')
      } else if x < 23 {
        console.log('medium')
      } else {
        console.log('large')
      }
      `
      const output = compile(input)
      expect(output).toBe(`if (x < 5) {
        console.log('small')
      } else if (x < 23) {
        console.log('medium')
      } else {
        console.log('large')
      }`)
    })
    it('if expressions with single statements', () => {
      const input = 'const foo = if x < 5 { "small" } else { "medium" }'
      const output = compile(input)
      expect(output).toBe('const foo = (x < 5 ? "small" : "medium")')
    })
    it('if expressions with multiple statements', () => {
      const input = `const foo = if x < 5 {
        const y = x*3
        y
      } else {
        const y = x/3
        console.log(y)
        y
      }`
      const output = compile(input)
      expect(output).toBe(`const foo = (x < 5 ? (() => { const y = x*3; return y; })() : (() => { const y = x/3; console.log(y); return y; })())`)
    })
    it('if expressions with mixed statements', () => {
      const input = `const foo = if x < 5 {
        const y = x*3
        y
      } else {
        42
      }`
      const output = compile(input)
      expect(output).toBe(`const foo = (x < 5 ? (() => { const y = x*3; return y; })() : 42)`)
    })
  })
})
